"""Scan an SBOM using the Trivy CLI.

--------------------------------------------------------------------------------
SPDX-FileCopyrightText: Copyright © 2022 Lockheed Martin <open.source@lmco.com>
SPDX-FileName: hopprcop/trivy/trivy_scanner.py
SPDX-FileType: SOURCE
SPDX-License-Identifier: MIT
--------------------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
--------------------------------------------------------------------------------
"""

from __future__ import annotations

import contextlib
import json
import os
import sys
import tempfile

from pathlib import Path
from subprocess import PIPE, Popen
from typing import TYPE_CHECKING, ClassVar

import rich

from hoppr import Affect, Component, ComponentType, HopprError, Metadata, Sbom, Vulnerability
from hoppr.utils import get_package_url
from typing_extensions import deprecated

from hopprcop.utils import build_bom_dict_from_purls, purl_check, unsupported_purl_feedback
from hopprcop.vulnerability_scanner import VulnerabilitySuper


if TYPE_CHECKING:
    from packageurl import PackageURL


class TrivyScanner(VulnerabilitySuper, author="Aquasec", name="Trivy", offline_mode_supported=True):  # pragma: no cover
    """Interacts with the trivy cli to scan an SBOM."""

    # used to store the operating system component discovered in the provided SBOM for generating the SBOM for trivy
    _os_component: Component | None = None

    trivy_os_distro = None

    required_tools_on_path: ClassVar[list[str]] = ["trivy"]
    supported_types: ClassVar[list[str]] = [
        "cargo",
        "conan",
        "deb",
        "gem",
        "golang",
        "gradle",
        "maven",
        "npm",
        "nuget",
        "pypi",
        "rpm",
    ]

    def __init__(self, offline_mode: bool = False, trivy_os_distro: str | None = None):
        self.offline_mode = offline_mode
        self.trivy_os_distro = trivy_os_distro or os.getenv("OS_DISTRIBUTION", None)
        super().__init__()

    def get_vulnerability_db(self) -> bool:
        """Downloads vulnerability database.

        Returns a boolean representation of success.
        """
        success = False
        dbs = {"database": "--download-db-only", "java database": "--download-java-db-only"}
        for db, flag in dbs.items():
            args = ["trivy", "image", flag, "--no-progress"]
            with Popen(args, stdout=PIPE, stdin=PIPE, stderr=PIPE) as process:
                # Trivy returns all messages as stderr in the 0.48.0 version.
                # To properly report the status of the database update the return code
                # is the primary check with scraping stderr being secondary.
                stdout, stderr = process.communicate()

                if process.returncode == 0:
                    if b"Downloading" in stderr:
                        rich.print(f"TrivyScanner: {db} downloaded")
                    else:
                        rich.print(f"TrivyScanner: No update needed for {db}")

                    success = True
                else:
                    rich.print(f"[yellow]TrivyScanner: Error occurred while updating the {db}[/]")
                    rich.print(f"[yellow]TrivyScanner: {stderr.decode()}[/]")
        return success

    def get_vulnerabilities_for_purl(self, purls: list[str]) -> list[Vulnerability]:
        """Get the vulnerabilities for a list of package URLS (purls).

        Returns a list of CycloneDX vulnerabilities.
        """
        bom = build_bom_dict_from_purls([get_package_url(purl) for purl in purls])
        self._add_operating_system_component(bom)

        return self.get_vulnerabilities_for_sbom(Sbom.parse_obj(bom))

    def get_vulnerabilities_for_sbom(self, bom: Sbom) -> list[Vulnerability]:
        """Get the vulnerabilities for a CycloneDx compatible Software Bill of Materials (SBOM).

        Returns a list of CycloneDX vulnerabilities.
        """
        # sourcery skip: low-code-quality
        # Need to give this method a through review in the future and potentially split complexity out
        results: list[Vulnerability] = []

        _win32 = sys.platform == "win32"
        with tempfile.NamedTemporaryFile(mode="w+", delete=not _win32, encoding="utf-8") as bom_file:
            # Remove tools from metadata due to cyclonedx-go only having partial support for spec version 1.5 (as of 0.72.0)
            # Add root component if it doesn't exist since trivy expects for this to be populated
            # TODO: roll back once full support is in cyclonedx-go
            copied_bom = bom.copy(deep=True)
            copied_bom.metadata = copied_bom.metadata or Metadata()
            copied_bom.metadata.tools = None
            copied_bom.metadata.component = copied_bom.metadata.component or Component(
                type=ComponentType.LIBRARY, name="trivycomponent"
            )

            # Trivy treats file type components as applications
            if copied_bom.metadata.component.type == ComponentType.FILE.value:
                copied_bom.metadata.component.type = ComponentType.APPLICATION

            bom_file.write(copied_bom.json())
            bom_file.flush()

            args = ["trivy", "sbom", "--scanners", "vuln", "--format", "cyclonedx", str(bom_file.name)]
            if self.offline_mode:
                args = [*args, "--skip-db-update", "--skip-java-db-update", "--offline-scan"]
            cache = os.getenv("CACHE_DIR")

            # Check for unsupported purl types and report if found
            purls = [component.purl for component in bom.components or [] if component.purl]
            unsupported_purl_feedback(
                self._scanner_name, self.supported_types, [get_package_url(purl) for purl in purls]
            )

            if cache is not None:
                args += ["--cache-dir", cache]

            with Popen(args, stdout=PIPE, stdin=PIPE, stderr=PIPE, text=True) as process:
                stdout, stderr = process.communicate()
                if process.returncode != 0:
                    raise HopprError(f"{self.__class__.__name__} generated an exception: {stderr}")

                if _win32:
                    bom_file.close()
                    Path(bom_file.name).unlink()

            bom_dict = json.loads(stdout)
            purl_by_bom_ref: dict[str, str] = {}
            for component in bom_dict.get("components", []):
                # Component does not have bom_ref or purl, no need for lookup since it would fail
                with contextlib.suppress(KeyError):
                    purl_by_bom_ref[component["bom-ref"]] = component["purl"]

            for vulnerability_dict in bom_dict.get("vulnerabilities", []):
                vulnerability = Vulnerability.parse_obj(vulnerability_dict)
                vulnerability.tools = self.scanner_tools()
                affects: list[Affect] = []
                affects.extend(vulnerability.affects)
                vulnerability.affects = []
                for affect in affects:
                    trivy_purl: str | None = str(affect.ref)
                    if trivy_purl and not trivy_purl.startswith("pkg"):
                        trivy_purl = purl_by_bom_ref.get(affect.ref)
                    if trivy_purl is None:
                        rich.print(f"[yellow]Could not find purl for ref {affect.ref}!")
                    else:
                        for purl in purls:
                            if purl_check(trivy_purl, purl):
                                new_affect = Affect.parse_obj({"ref": purl, "versions": affect.versions})
                                vulnerability.affects.append(new_affect)

                if vulnerability.ratings is not None:
                    results.append(vulnerability)

        return results

    @deprecated("TrivyScanner.get_vulnerabilities_by_purl is deprecated, use TrivyScanner.get_vulnerabilities_for_purl")
    def get_vulnerabilities_by_purl(self, purls: list[PackageURL]) -> dict[str, list[Vulnerability]]:
        """Get the vulnerabilities for a list of package URLS (purls).

        Returns a dictionary of package URL to vulnerabilities or none if no vulnerabilities are found.
        """
        bom = build_bom_dict_from_purls(purls)
        self._add_operating_system_component(bom)

        return self.get_vulnerabilities_by_sbom(Sbom.parse_obj(bom))

    @deprecated("TrivyScanner.get_vulnerabilities_by_sbom is deprecated, use TrivyScanner.get_vulnerabilities_for_sbom")
    def get_vulnerabilities_by_sbom(self, bom: Sbom) -> dict[str, list[Vulnerability]]:
        """Accepts a cyclone dx compatible SBOM and returns a list of vulnerabilities.

        Returns a dictionary of package URL to vulnerabilities or none if no vulnerabilities are found.
        """
        results: dict[str, list[Vulnerability]] = {}

        _win32 = sys.platform == "win32"
        with tempfile.NamedTemporaryFile(mode="w+", delete=not _win32, encoding="utf-8") as bom_file:
            # Remove tools from metadata due to cyclonedx-go only having partial support for spec version 1.5 (as of 0.72.0)
            # Add root component if it doesn't exist since trivy expects for this to be populated
            # TODO: roll back once full support is in cyclonedx-go
            copied_bom = bom.copy(deep=True)
            copied_bom.metadata = copied_bom.metadata or Metadata()
            copied_bom.metadata.tools = None
            copied_bom.metadata.component = copied_bom.metadata.component or Component(
                type=ComponentType.LIBRARY, name="trivycomponent"
            )

            bom_file.write(copied_bom.json())
            bom_file.flush()

            args = ["trivy", "sbom", "--scanners", "vuln", "--format", "cyclonedx", str(bom_file.name)]
            if self.offline_mode:
                args = [*args, "--skip-db-update", "--skip-java-db-update", "--offline-scan"]
            cache = os.getenv("CACHE_DIR")

            # Check for unsupported purl types and report if found
            purls = [get_package_url(component.purl) for component in bom.components or [] if component.purl]
            unsupported_purl_feedback(self._scanner_name, self.supported_types, purls)

            if cache is not None:
                args += ["--cache-dir", cache]

            with Popen(args, stdout=PIPE, stdin=PIPE, stderr=PIPE, text=True) as process:
                stdout, stderr = process.communicate()
                if not stdout and stderr:
                    raise HopprError(f"{self.__class__.__name__} generated an exception: {stderr}")

                if _win32:
                    bom_file.close()
                    Path(bom_file.name).unlink()

            bom_dict = json.loads(stdout)

            for vuln_dict in bom_dict["vulnerabilities"]:
                vuln = Vulnerability.parse_obj(vuln_dict)
                for affect in vuln.affects:
                    *_, purl_str = str(affect.ref).split("#")
                    affect.ref = purl_str or affect.ref

                    if vuln.ratings is not None:
                        results.setdefault(affect.ref, [])
                        results[affect.ref].append(vuln)

                vuln.tools = self.scanner_tools()

        return results

    def _add_operating_system_component(self, bom: dict):
        version = None
        distro = None

        if self.trivy_os_distro is not None:
            parts = self.trivy_os_distro.split(":")

            if len(parts) != 2:
                rich.print(f"{self.trivy_os_distro} is an invalid distribution ")
            else:
                distro = parts[0]
                version = parts[1]
        elif self._os_component is not None:
            version = self._os_component.version
            distro = self._os_component.name

        if version is not None and distro is not None:
            component = {
                "bom-ref": "ab16d2bb-90f7-4049-96ce-8c473ba13bd2",
                "type": "operating-system",
                "name": distro,
                "version": version,
            }
            bom["components"].append(component)



"""Documents an assessment of impact and exploitability of vulnerabilities.

--------------------------------------------------------------------------------
SPDX-FileCopyrightText: Copyright © 2022 Lockheed Martin <open.source@lmco.com>
SPDX-FileName: hopprcop/enhancements/analysis_assessment.py
SPDX-FileType: SOURCE
SPDX-License-Identifier: MIT
--------------------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
--------------------------------------------------------------------------------
"""

from __future__ import annotations

from datetime import date, datetime
from pathlib import Path
from typing import TYPE_CHECKING, ClassVar

import rich

from hoppr import Affect, HopprLoadDataError, cdx, utils
from hoppr.utils import get_package_url
from typing_extensions import deprecated


if TYPE_CHECKING:
    from hoppr import Vulnerability
    from packageurl import PackageURL


class Analysis:
    """Documents an assessment of impact and exploitability."""

    enabled: bool = False
    assessments: ClassVar[list[Assessment]] = []
    _assessment_types: ClassVar[list[str]] = []
    _assessment_findings: ClassVar[list[str]] = []
    _assessment_descriptions: ClassVar[list[str]] = []

    class Assessment:
        assessment_date: date | None
        vulnerability: str | None
        state: str | None
        justification: str | None
        package: Package
        response: list[cdx.ResponseEnum] | None = None
        detail: str | None

        def __init__(self, assessment: dict):
            # Ref: https://cyclonedx.org/docs/1.5/json/#vulnerabilities_items_analysis
            valid_state: list[str] = [
                "resolved",
                "resolved_with_pedigree",
                "exploitable",
                "in_triage",
                "false_positive",
                "not_affected",
            ]
            valid_justification: list[str] = [
                "",
                "code_not_present",
                "code_not_reachable",
                "requires_configuration",
                "requires_dependency",
                "requires_environment",
                "protected_by_compiler",
                "protected_at_runtime",
                "protected_at_perimeter",
                "protected_by_mitigating_control",
            ]

            # Required Properties from analysis.assessment.yml
            self.assessment_date = assessment.get("assessmentDate")
            if not self.assessment_date:
                raise KeyError("assessments.assessmentDate must the date the assessment conducted.")

            self.vulnerability = assessment.get("vulnerability")
            if not self.vulnerability:
                raise KeyError("assessments.vulnerability must contain a vulnerability id")

            self.state = assessment.get("state")
            if self.state not in valid_state:
                raise KeyError(f"assessments.state must be one of {valid_state}")

            self.justification = assessment.get("justification") or ""
            if self.justification not in valid_justification:
                raise KeyError(f"assessments.justification must be one of {valid_justification[1:]}")

            if self.state in ["not_affected", "false_positive"] and not self.justification:
                # justification: required if state=not_affected or false_positive
                raise KeyError("assessments.justification is required when state=not_affected or false_positive")

            self.package = self.Package(assessment.get("package") or {})

            # Optional Properties
            if "response" in assessment:
                valid_response: list[str] = [
                    "can_not_fix",
                    "will_not_fix",
                    "update",
                    "rollback",
                    "workaround_available",
                ]
                if assessment.get("response") in valid_response:
                    self.response = [cdx.ResponseEnum(assessment.get("response"))]
                else:
                    raise KeyError(f"assessments.response must be one of {valid_response}")

            self.detail = assessment.get("detail") or ""
            if not self.detail and self.state not in ["exploitable", "in_triage"]:
                raise KeyError("assessments.detail must be set when state is not 'exploitable' or 'in_triage'")

        class Package:
            type: str | None
            name: str | None
            version: str | None

            def __init__(self, package: dict):
                valid_packages: list[str] = [
                    "cargo",
                    "composer",
                    "conan",
                    "deb",
                    "docker",
                    "gem",
                    "golang",
                    "gradle",
                    "maven",
                    "npm",
                    "nuget",
                    "oci",
                    "pypi",
                    "rpm",
                ]

                self.type = package.get("type")
                if self.type not in valid_packages:
                    raise KeyError(f"assessments.package.type must be one of {valid_packages}")

                self.name = package.get("name")
                if not self.name:
                    raise KeyError("assessments.package.name must contain a valid package name")

                self.version = package.get("version")
                if not self.version:
                    raise KeyError("assessments.package.version must contain a valid version")

    def _set_assessment_path(self, path: str | Path) -> bool:  # pragma: no cover
        """Sets the path to the analysis.assessment.yml file.

        Argument path can be relative or fully qualified path
        For example "./" will look for analysis.assessment.yml in the current directory
        """
        assessment_file_path: Path = Path(path) / "analysis.assessment.yml"

        # Clears lists if users are scanning multiple SBOMs
        self.assessments.clear()
        self._assessment_types.clear()
        self._assessment_findings.clear()
        self._assessment_descriptions.clear()

        if not assessment_file_path.exists():
            rich.print(f"Oops, analysis.assessment.yml doesn't exist in: {assessment_file_path.absolute()}")
            return False

        try:
            self.enabled = True
            assessment_dict = utils.load_file(assessment_file_path)
            if not isinstance(assessment_dict, dict):
                raise HopprLoadDataError("Pulled artifact data was not loaded as dictionary")
            for assessment in assessment_dict["assessments"]:
                assessment_obj = self.Assessment(assessment)
                if assessment_obj.package.type and assessment_obj.package.type not in self._assessment_types:
                    self._assessment_types.append(assessment_obj.package.type)
                self.assessments.append(assessment_obj)
            rich.print(f"Loaded: {assessment_file_path}")
        except KeyError as exc:
            raise KeyError(f"Invalid value specified in analysis.assessment.yml: {exc}") from None
        except Exception as exc:  # pylint: disable=broad-except
            rich.print(f"Failed to Load: {assessment_file_path}")
            rich.print(f"{self.__class__.__name__} generated an Assessment exception: {exc}")
            return False
        return True

    def apply_assessments(self, scanner_results: list[Vulnerability]):
        """Checks scanner results against analysis.assessment.yml.

        Returns a dictionary of vulnerabilities with related analysis assessment added if applicable.
        """
        for vuln in scanner_results:
            try:
                for assessment in self.assessments:
                    self._determine_assessment(assessment, vuln)

            except Exception as exc:  # pylint: disable=broad-except
                rich.print(
                    f"{vuln.id} and {assessment.package.type}/{assessment.package.name}:{assessment.package.version} generated an Assessment exception: {exc}"
                )

    def _requires_justification(self, state: str | None) -> bool:
        return state in [cdx.ImpactAnalysisState.NOT_AFFECTED.value, cdx.ImpactAnalysisState.FALSE_POSITIVE.value]

    def _determine_assessment(self, assessment_obj: Assessment, vuln: Vulnerability):
        """Determines if purl matches an assessment assertion. Documents analysis.

        Updates description. Description is reflected in enhanced json and html reports
        """
        if (
            not self._package_is_affected_by_vuln(assessment_obj.package, vuln.affects)
            or not vuln.ratings
            or assessment_obj.vulnerability != vuln.id
        ):
            return

        average_severity: str = self._get_average_severity(vuln.ratings)
        for rating in vuln.ratings:
            self._check_rating(assessment_obj, average_severity, vuln, rating)

        vuln.analysis = cdx.Analysis(
            state=cdx.ImpactAnalysisState(assessment_obj.state),
            justification=(
                cdx.ImpactAnalysisJustification(assessment_obj.justification)
                if assessment_obj.justification or self._requires_justification(assessment_obj.state)
                else None
            ),
            response=assessment_obj.response,
            detail=assessment_obj.detail,
            firstIssued=datetime.combine(assessment_obj.assessment_date or date.min, datetime.min.time()),
            lastUpdated=datetime.now(),
        )

        response_str = "" if assessment_obj.response is None else str(assessment_obj.response[0])
        description = f"<b class='font-medium'>Initial Severity:</b> {average_severity}<br><b class='font-medium'>Description</b>: VULN_DESCRIPTION<br><b class='font-medium'>Assessment of Impact:</b> {assessment_obj.assessment_date}<br><b class='font-medium'>State:</b> {assessment_obj.state}<br><b class='font-medium'>Justification:</b> {assessment_obj.justification}<br><b class='font-medium'>Response:</b> {response_str}<br><b class='font-medium'>Detail:</b> {assessment_obj.detail}<br><b class='font-medium'>Scan Date:</b> {datetime.now().strftime('%Y-%m-%d')}"
        desc_key = f"{vuln.id}:{description}"
        if desc_key not in self._assessment_descriptions:
            self._assessment_descriptions.append(desc_key)
            vuln.description = description.replace("VULN_DESCRIPTION", str(vuln.description))

    def _package_is_affected_by_vuln(self, assessment_package: Assessment.Package, affects: list[Affect]) -> bool:
        """Determines if the package from the assessment is in the affects list of a vulnerability.

        Returns a bool of whether the package is in the affects list.
        """
        return (
            len(
                [
                    purl
                    for purl in [get_package_url(affect.ref) for affect in affects]
                    if assessment_package.type == purl.type
                    and assessment_package.name == purl.name
                    and (assessment_package.version or "") == (purl.version or "")
                ]
            )
            > 0
        )

    @deprecated("Analysis.check_assessments is deprecated, use Analysis.apply_assessments")
    def check_assessments(
        self, scanner_results: dict[str, list[Vulnerability]], scanner_name: str
    ) -> dict[str, list[Vulnerability]]:
        """Checks scanner results against analysis.assessment.yml.

        Returns a dictionary of vulnerabilities with related analysis assessment added if applicable.
        """
        results: dict = {}
        scanner = scanner_name.lower().replace("scanner", "")
        for purl, vuln_list in scanner_results.items():
            try:
                purl_data = get_package_url(purl)
                if len(vuln_list) != 0 or purl_data.type in self._assessment_types:
                    for assessment in self.assessments:
                        self._check_assessment(purl_data, assessment, vuln_list)
                results[purl] = vuln_list

            except Exception as exc:  # pylint: disable=broad-except
                results[purl] = vuln_list
                rich.print(f"{scanner} generated an Assessment exception: {exc}")

        return results

    @deprecated("Analysis._check_assessment is deprecated, use Analysis._determine_assessment")
    def _check_assessment(self, purl_data: PackageURL, assessment_obj: Assessment, vuln_list: list[Vulnerability]):
        """Determines if purl matches an assessment assertion. Documents analysis.

        Updates description. Description is reflected in enhanced json and html reports
        """
        if assessment_obj.package.type != purl_data.type or assessment_obj.package.name != purl_data.name:
            return

        for vuln in vuln_list:
            average_severity: str = self._get_average_severity(vuln.ratings)
            if (
                vuln.ratings
                and assessment_obj.vulnerability == vuln.id
                and purl_data.version == assessment_obj.package.version
            ):
                for rating in vuln.ratings:
                    self._check_rating(assessment_obj, average_severity, vuln, rating)
                vuln.analysis = cdx.Analysis(
                    state=cdx.ImpactAnalysisState(assessment_obj.state),
                    justification=cdx.ImpactAnalysisJustification(assessment_obj.justification),
                    response=assessment_obj.response,
                    detail=assessment_obj.detail,
                    firstIssued=datetime.combine(assessment_obj.assessment_date or date.min, datetime.min.time()),
                    lastUpdated=datetime.now(),
                )
                response_str = "" if assessment_obj.response is None else str(assessment_obj.response[0])
                description = f"<b class='font-medium'>Initial Severity:</b> {average_severity}<br><b class='font-medium'>Description</b>: VULN_DESCRIPTION<br><b class='font-medium'>Assessment of Impact:</b> {assessment_obj.assessment_date}<br><b class='font-medium'>State:</b> {assessment_obj.state}<br><b class='font-medium'>Justification:</b> {assessment_obj.justification}<br><b class='font-medium'>Response:</b> {response_str}<br><b class='font-medium'>Detail:</b> {assessment_obj.detail}<br><b class='font-medium'>Scan Date:</b> {datetime.now().strftime('%Y-%m-%d')}"
                desc_key = f"{vuln.id}:{description}"
                if desc_key not in self._assessment_descriptions:
                    self._assessment_descriptions.append(desc_key)
                    vuln.description = description.replace("VULN_DESCRIPTION", str(vuln.description))

    def _check_rating(self, assessment_obj: Assessment, average_severity: str, vuln: Vulnerability, rating: cdx.Rating):
        """Determines if a rating adjustment is needed based on the analysis assessment.

        Documents rating adjustment if applicable.
        """
        adjusted_severity: cdx.Severity | None = None
        initial_severity = rating.severity

        if assessment_obj.state not in [] and average_severity not in ["info", "none", "unknown"]:
            rating.severity = cdx.Severity.low
            adjusted_severity = rating.severity

        assessment_message = f"[bold yellow]{vuln.id}[/bold yellow] severity:'INITIAL_SEVERITY' changed to 'NEW_SEVERITY' based Assessment of Impact state : {assessment_obj.state}. Ref: {assessment_obj.package.type}:{assessment_obj.package.name}:{assessment_obj.package.version} in analysis.assessment.yml"

        if assessment_message not in self._assessment_findings:
            self._assessment_findings.append(assessment_message)
            if average_severity == rating.severity:
                rich.print(
                    assessment_message.replace("INITIAL_SEVERITY", average_severity).replace(
                        "changed to 'NEW_SEVERITY'", "was not changed"
                    )
                )
            else:
                rich.print(
                    assessment_message.replace("INITIAL_SEVERITY", average_severity).replace(
                        "NEW_SEVERITY", str(rating.severity)
                    )
                )

        if adjusted_severity is not None:
            assessment_obj.detail = f"Severity of finding reduced from {initial_severity} to {adjusted_severity} due to analysis asserted by user in analysis.assessment.yml. Detail: {assessment_obj.detail}"
            if not rating.justification and assessment_obj.justification:
                rating.justification = assessment_obj.justification

    def _get_average_severity(self, ratings: list[cdx.Rating] | None) -> str:
        """Gets the average severity from the vuln.ratings.

        Returns str
        """
        severities = ["info", "low", "medium", "high", "critical"]

        if not isinstance(ratings, list):
            raise TypeError("Ratings was not passed as a list of Rating") from None
        indexes = [
            severities.index(str(rating.severity))
            for rating in ratings
            if rating.severity and str(rating.severity) in severities
        ]
        return severities[int(sum(indexes) / len(indexes))]



# Copyright 2022-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import logging
import os
import sys
import time
from datetime import datetime, timedelta
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import requests
import yaml

from packageurl import PackageURL
from tinydb import TinyDB, Query
from tinydb.table import Document

from .exception import AccessDeniedException, RateLimitException
from .model import OssIndexComponent
from .serializer import json_decoder, OssIndexJsonEncoder

logger = logging.getLogger("ossindex")

if sys.version_info >= (3, 8):
    from importlib.metadata import version as meta_version
else:
    from importlib_metadata import version as meta_version

ossindex_lib_version: str = "TBC"
try:
    ossindex_lib_version = str(meta_version("ossindex-lib"))
except Exception:
    ossindex_lib_version = "DEVELOPMENT"


class OssIndex:
    DEFAULT_CONFIG_FILE = ".oss-index.config"

    _caching_enabled: bool = False
    _cache_directory: str = ".ossindex"
    _cache_ttl_in_hours: int = 12

    _oss_index_api_version: str = "v3"
    _oss_index_host: str = "https://ossindex.sonatype.org"
    _oss_max_coordinates_per_request: int = 128
    oss_index_authentication: Optional[requests.auth.HTTPBasicAuth] = None

    def __init__(self, *, enable_cache: bool = True, cache_location: Optional[str] = None) -> None:
        self._caching_enabled = enable_cache
        if self._caching_enabled:
            logger.info("OssIndex caching is ENABLED")
            self._setup_cache(cache_location=cache_location)
        self._attempt_config_load()

    def has_ossindex_authentication(self) -> bool:
        return self.oss_index_authentication is not None

    def _attempt_config_load(self) -> None:
        try:
            config_filename: str = os.path.join(os.path.expanduser("~"), OssIndex.DEFAULT_CONFIG_FILE)
            with open(config_filename, "r") as ossindex_confg_f:
                ossindex_config = yaml.safe_load(ossindex_confg_f.read())
                self.oss_index_authentication = requests.auth.HTTPBasicAuth(
                    ossindex_config["username"], ossindex_config["password"]
                )
        except FileNotFoundError:
            pass

    def get_component_report(self, packages: List[PackageURL]) -> List[OssIndexComponent]:
        logger.debug(f"A total of {len(packages)} Packages to be queried against OSS Index")
        return self._get_results(packages=packages)

    def purge_local_cache(self) -> None:
        if self._caching_enabled:
            logger.info("Truncating local cache database as requested")
            with self._get_cache_db() as db:
                db.truncate()
                logger.info("Local OSS Index cache has been purged")

    def _chunk_packages_for_oss_index(self, packages: List[PackageURL]) -> List[List[PackageURL]]:
        """
        Splits up the list of packages into lists that are of a size consumable by OSS Index
        APIs.

        :param packages: List[PackageURL]
        :return: List[List[PackageURL]]
        """
        return [
            packages[i : i + self._oss_max_coordinates_per_request]
            for i in range(0, len(packages), self._oss_max_coordinates_per_request)
        ]

    def _get_api_url(self, api_uri: str) -> str:
        return f"{self._oss_index_host}/api/{self._oss_index_api_version}/{api_uri}"

    def _get_cached_results(self, packages: List[PackageURL]) -> Tuple[List[PackageURL], List[OssIndexComponent]]:
        """
        Takes a list of packages and returns two Lists:
            1. Packages without cached results
            2. Cached results for those packages where they exist

        :param packages: List[PackageURL]
        :return: (List[PackageURL], List[OssIndexComponent])
        """
        if not self._caching_enabled:
            # This should not be possible, but adding for developer safety
            return packages, []

        now = datetime.now()
        cached_results: List[OssIndexComponent] = []
        non_cached_packaged: List[PackageURL] = []
        with self._get_cache_db() as cache_db:
            for package in packages:
                logger.debug(f"   Checking in cache for {package.to_string()}")
                cache_results: List[Document] = cache_db.search(Query().coordinates == package.to_string())
                if not cache_results:
                    # Not cached
                    logger.debug("      Not in cache")
                    non_cached_packaged.append(package)
                elif datetime.strptime(cache_results[0]["expiry"], "%Y-%m-%dT%H:%M:%S.%f") < now:
                    logger.debug("      Cached, but result expired")
                    non_cached_packaged.append(package)
                else:
                    logger.debug("      Cached, loading from cache")
                    cached_results.append(
                        json.loads(
                            f'[{cache_results[0]["response"]}]',
                            object_hook=json_decoder,
                        ).pop()
                    )

        return non_cached_packaged, cached_results

    @staticmethod
    def _get_headers() -> Dict[str, str]:
        return {
            "Accept": "application/vnd.ossindex.component-report.v1+json",
            "Content-type": "application/vnd.ossindex.component-report-request.v1+json",
            "User-Agent": f"python-oss-index-lib@{ossindex_lib_version}",
        }

    def _get_results(self, packages: List[PackageURL]) -> List[OssIndexComponent]:
        results: List[OssIndexComponent] = []

        # First get any cached results
        if self._caching_enabled:
            logger.debug("Checking local cache for any usable results...")
            packages, results = self._get_cached_results(packages=packages)
            logger.debug(f"   {len(results)} cached results found leaving {len(packages)} to ask OSS Index")

        # Second, chunk up packages for which we have no cached results and query OSS Index
        chunk: List[PackageURL]
        chunks = self._chunk_packages_for_oss_index(packages=packages)
        logger.debug(f"Split {len(packages)} packages into {len(chunks)} chunks for OSS requests")
        for chunk in chunks:
            logger.debug("  Getting chunk results from OSS Index...")
            results = results + self._make_oss_index_component_report_call(packages=chunk)

        logger.debug(f"Total of {len(results)} results (including cached)")
        return results

    def _make_oss_index_component_report_call(self, packages: List[PackageURL]) -> List[OssIndexComponent]:
        response = requests.post(
            url=self._get_api_url("component-report"),
            headers=self._get_headers(),
            json={"coordinates": list(map(lambda p: str(p.to_string()), packages))},
            auth=self.oss_index_authentication,
        )

        if response.status_code != 200:
            print(response.status_code)
            if response.status_code == 429:
                raise RateLimitException()

        results: List[OssIndexComponent] = list(response.json(object_hook=json_decoder))
        if self._caching_enabled:
            self._upsert_cache_with_oss_index_responses(oss_components=results)
        return results

    def _upsert_cache_with_oss_index_responses(self, oss_components: List[OssIndexComponent]) -> None:
        if not self._caching_enabled:
            return

        now = datetime.now()
        cache_expiry = now + timedelta(hours=self._cache_ttl_in_hours)
        oc: OssIndexComponent
        with self._get_cache_db() as cache_db:
            for oc in oss_components:
                query: Query = Query()
                if cache_query_result := cache_db.search(query.coordinates == oc.coordinates):
                    # Update existing cache
                    logger.debug(f"    Might refresh cache for {oc.coordinates}")
                    if now > datetime.strptime(cache_query_result[0]["expiry"], "%Y-%m-%dT%H:%M:%S.%f"):
                        # Cache expired - update it!
                        logger.debug(f"        Cache expired for {oc.coordinates} - UPDATING CACHE")
                        cache_db.update(
                            {
                                "response": json.dumps(oc, cls=OssIndexJsonEncoder),
                                "expiry": cache_expiry.isoformat(),
                            },
                            query.coordinates == oc.coordinates,
                        )
                    else:
                        logger.debug(f"    Cache is still valid for {oc.coordinates} - not updating")

                else:
                    # New component for caching
                    logger.debug(f"    Caching new Component results for {oc.coordinates}")
                    cache_db.insert(
                        {
                            "coordinates": oc.coordinates,
                            "response": json.dumps(oc, cls=OssIndexJsonEncoder),
                            "expiry": cache_expiry.isoformat(),
                        }
                    )

    def _get_cache_db(self) -> TinyDB:
        return TinyDB(os.path.join(self._cache_directory, "ossindex.json"))

    def _setup_cache(self, cache_location: Optional[str] = None) -> None:
        full_cache_path: str
        if not cache_location:
            full_cache_path = os.path.join(Path.home(), self._cache_directory)
        else:
            full_cache_path = os.path.join(cache_location, self._cache_directory)

        if not os.path.exists(full_cache_path):
            Path(full_cache_path).mkdir(parents=True, exist_ok=True)

        self._cache_directory = str(Path(full_cache_path))



"""EPSS API results model."""

from __future__ import annotations

from typing import Annotated

from pydantic import BaseModel, Field


class Datum(BaseModel):
    """Model to describe EPSS datum."""

    cve: str
    epss: str
    percentile: str
    date: str | None = None


class EPSSResult(BaseModel):
    """Model to describe an EPSS result."""

    status: str
    status_code: Annotated[int, Field(alias="status-code")]
    version: str
    access: str | None = None
    total: int | None = None
    offset: int | None = None
    limit: int | None = None
    data: list[Datum]
